/*
 * common.c - shared system foundation functions
 *   by uratan! 2020.6.23
 */
#include "common.h"
#include <chip.h>

	/*
	 * functions prototpes are in common.h
	 */

#if gUDEBUG
/*
 * table for bit reverse
 */
static const u_char rev_bits[];
#endif /* gUDEBUG */

/* =============================================================== */

/*
 * global lock
 */
u_long IntDisable(void)
{
	u_long imf;

	imf = __get_PRIMASK();		/* imf: 1: disabled  0: enabled */
	__disable_irq();
	return imf;
}

/*
 * recover lock state
 */
void IntRestore(u_long imf)
{
	if(imf == 0) {
		__enable_irq();
	}
}

/* =============================================================== */

#if gUDEBUG
/*
 * initial SSPn for UART use
 */
void cm_ssp_uart_init(u_long baudrate, u_long mainCLK)
{
	u_long Fdiv;

	/*
	 * confirm peripheral reset
	 */
	LPC_SYSCTL->PRESETCTRL &= ~(1 << RESET_SSPn);

	/*
	 * enable clock for SSPn
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_SSPn);

	/*
	 * enable transfer clock
	 *   12MHz / (13 * 4) / 4 = 57692.3076
	 */
	LPC_SYSCTL->SSPnCLKDIV = 52;		/* max 1/255 */
			/* tune this register for 1/4 in power_down() */

	/*
	 * release peripheral reset
	 */
	LPC_SYSCTL->PRESETCTRL |= (1 << RESET_SSPn);

	/*
	 * setup SSPn for uart use
	 */
		/*
		 * uart format is: Startbit + 8bit(from LSB) + stopbit
		 *
		 *      pre-drive (half-clk)                     post-drive (half-clk)
		 *         |                                         |
		 *         V                                         V
		 *          |<-------------- 10bit ---------------->|
		 *  Hi-Z  | |                                       | |  Hi-Z
		 * =======+-+   +---+---+---+---+---+---+---+---+---+-+=======
		 *   ...-->s| S |LSB|   |   |   |   |   |   |MSB| s->s
		 *          +---+---+---+---+---+---+---+---+---+
		 */
	Fdiv = mainCLK / LPC_SYSCTL->SSPnCLKDIV / baudrate;
	LPC_SSPn->CPSR = Fdiv;		/* clock 1/4 here */
	LPC_SSPn->CR0 = 0x0019;		/* 1/1, TI, 10bit */
	LPC_SSPn->CR1 = SSP_CR1_MASTER_EN | SSP_CR1_SSP_EN;
	LPC_SSPn->IMSC = 0x0000;	/* no use interrupt */

	/*
	 * before drive real pin, initialize MOSI0 output state to high
	 *   (to keep above pre-drive high)
	 */
	LPC_SSPn->DR = 0xffff;
		/* wait until the data is completely sent */
	while(LPC_SSPn->SR & SSP_STAT_BSY) {
		;
	}

	/*
	 * connect SSPn/MOSIn to the pin
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

#if USE_SSPn == 0
	LPC_IOCON->PIO0[9] = 0x0091;		/* select MOSI0, with pull-up-R */
#else /* USE_SSPn == 0 */
	LPC_IOCON->PIO0[21] = 0x0092;		/* select MOSI1, with pull-up-R */
#endif /* USE_SSPn == 0 */

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);
}

/*
 * send a char thru SSPn
 */
void cm_putchar(int ch)
{
	u_long frame;

	/*
	 * construct uart frame
	 */
		//asm("	rbit r0,r0");			/* this core do not have RBIT... */
	frame =  rev_bits[(ch >> 4) & 0x0f] << 0;
	frame |= rev_bits[(ch >> 0) & 0x0f] << 4;
	frame = (frame << 1) | 0x0001;		/* add start/stop-bit */

	/*
	 * wait until the unit is idle
	 */
	while(LPC_SSPn->SR & SSP_STAT_BSY) {
		;
	}

	/*
	 * start to send
	 */
	LPC_SSPn->DR = frame;

	if(0) {
		/*
		 * TEST: how about repeated transfer ? --> no problem
		 */
		while(LPC_SSPn->SR & SSP_STAT_BSY) {
			;
		}
		LPC_SSPn->DR = frame;
	}

	return;
}
#endif /* gUDEBUG */

/* =============================================================== */

/*
 * is the char hexadecimal ?   (pass lower-case only)
 */
int is_hex(u_char c)
{
	if(('0' <= c) && (c <= '9'))
		return TRUE;
	if(('a' <= c) && (c <= 'f'))
		return TRUE;
	return FALSE;
}

/*
 * convert hexadecimal char to binary  (lower-case only)
 */
u_long to_bin(u_char c)
{
	if(('a' <= c) && (c <= 'f'))
		return(c - 'a' + 10);
	return(c - '0');
}

/* =============================================================== */

#if gUDEBUG
/*
 * table for bit reverse
 */
static const u_char rev_bits[] = {
/*0x00*/	0x00,
/*0x01*/	0x08,
/*0x02*/	0x04,
/*0x03*/	0x0c,
/*0x04*/	0x02,
/*0x05*/	0x0a,
/*0x06*/	0x06,
/*0x07*/	0x0e,
/*0x08*/	0x01,
/*0x09*/	0x09,
/*0x0a*/	0x05,
/*0x0b*/	0x0d,
/*0x0c*/	0x03,
/*0x0d*/	0x0b,
/*0x0e*/	0x07,
/*0x0f*/	0x0f,
};
#endif /* gUDEBUG */

/* =============================================================== */
