/*
 * main.c - main for LPC11U35
 *   by uratan! 2020.6.21
 */
//#define __MAIN_C__		/* <== inform common.h to alloc variables */
#include "common.h"
#include <chip.h>
	/*
	 * my LPC11U35FHI33/501 (HVQFN33)
	 *   *---------+
	 *   | 0 11U35 |
	 *   | Z 57 02 |
	 *   | 4492A50 | ... year 4?, week 49?, revision 'A' ??
	 *   +---------+
	 *      | Dev.ID:  0x0000BC40
	 *      | B.L.Ver:  7.2
	 */

/*
 * for uart-debug
 */
#define RPRINTF    if(gUDEBUG)rprintf

/*
 * SysTick base interval
 */
#define SYSTICK_Hz   1000		/* get 1msec interval */

/*
 * functions
 */
int main(void);
u_long initial_sysCLK(int use_PLL, u_long divider);
void control_USB_PLL(int start);
void configure_GPIO(void);
void initial_usb_clk_n_pins(void);
void init_SysTick(u_long sysCLK);
void SysTick_Handler(void);
void def_Exception_Handler(void);
void def_NMI_Handler(void);
void def_HardFault_Handler(void);
void def_SVCall_Handler(void);
void def_PendSV_Handler(void);

/*
 * my variables
 */
VOLATILE int GotSUD;	/* SetUp Data */
VOLATILE int GotCOD;	/* Control-OUT Data */
VOLATILE int Sleep;		/* shall Sleep becaue in Suspend mode */
VOLATILE int systick_raw;

/*
 * externs
 */
	/* in usart0.c */
extern void usart0_init(u_long baudrate, u_long mainCLK);
	/* in iap_fup.c */
extern void iap_fup_init(void);
extern u_long iap_Fcclk_kHz;			/* System Clock Frequency [kHz] */
	/* in pdown.c */
extern void pdown_init(void);
	/* in usbhw.c */
extern int uhw_Init(void);
extern void uhw_Connect(int conn);
	/* in usbcore.c */
extern void ucore_InitialCore(void);
extern void ucore_TryPowerDown(void);
extern void ucore_SetupCommand(void);
extern void ucore_ProceedSetupCommand(void);
	/* in periph.c */
extern void TD_Init(void);
extern void TD_Poll(void);
extern int TD_Suspend(void);
extern void TD_Resume(void);

#if gUDEBUG
	/* in dscr.h.S */
extern u_long *bcdDevice, *idVendor, *idProduct;
#endif /* gUDEBUG */

/* =============================================================== */

/*
 *
 */
int main(void)
{
	u_long sysCLK;
	u_long mainCLK;

	/*
	 * mask all interrupts by PRIMASK flag in the core
	 */
	__disable_irq();

	/*
	 * why bootROM enables these ?
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_SSP0);
	LPC_SYSCTL->SSP0CLKDIV = 0;		/* disable */
		// SYSCTL_CLOCK_USB and SYSCTL_CLOCK_USBRAM are also enabled

	/*
	 * select mainCLK as USB clock by USBCLKSEL
	 *   !! both mainCLK and USB PLL must be running to switch !!
	 */
	control_USB_PLL(/*start=*/ 1);
	LPC_SYSCTL->USBCLKDIV = 1;			/* 1/1 */
	LPC_SYSCTL->USBCLKSEL = 1;		/* select mainCLK */
	LPC_SYSCTL->USBCLKUEN = 0;			/* feed it */
	LPC_SYSCTL->USBCLKUEN = 1;
	control_USB_PLL(/*start=*/ 0);		/* do not need more */

	/*
	 * setup system clock for 48MHz
	 */
	sysCLK = initial_sysCLK(/*use_PLL=*/ 1, /*divider=*/ 1);
	mainCLK = sysCLK * LPC_SYSCTL->SYSAHBCLKDIV;
	iap_Fcclk_kHz = sysCLK / 1000;

	/*
	 * setup uart for debug, etc.
	 */
#if gUDEBUG
	cm_ssp_uart_init(57600, mainCLK);
	rprintf(
		"\r\n"
		"/*\r\n"
		" * start thumbUZK (%d+%dMHz)\r\n"
		" *   ver %04x\r\n"
		" * idVendor: 0x%04x  idProduct: 0x%04x\r\n"
		" */\r\n"
		, mainCLK / 1000 / 1000
		, sysCLK / 1000 / 1000
		, &bcdDevice
		, &idVendor, &idProduct
	);
//	rprintf("FLASHTIM is: 0x%08x\r\n", LPC_FMC->FLASHTIM);	/*0x4003_c010*/
//	rprintf("SYSAHBCLKDIV is: %d\r\n", LPC_SYSCTL->SYSAHBCLKDIV);
//	rprintf("SYSAHBCLKCTRL is: 0x%08x\r\n", LPC_SYSCTL->SYSAHBCLKCTRL);
#endif /* gUDEBUG */

#if gUDEBUG
if(0) {
				/* dump bootROM */
	u_char *mem;
	int i, j;

	mem = (u_char *)0x1fff0000;
	for(i=0; i<1024; i++) {			/* 16k */
		rprintf("%08x:", mem);
		for(j=0; j<16; j++) {
			rprintf(" %02x", *mem++);
		}
		rprintf("\r\n");
	}
}
if(0) {
				/* dump SRAM0 */
	u_char *mem;
	int i, j;

	mem = (u_char *)0x10000000;
	for(i=0; i<512; i++) {
		rprintf("%08x:", mem);
		for(j=0; j<16; j++) {
			rprintf(" %02x", *mem++);
		}
		rprintf("\r\n");
	}
}
#endif /* gUDEBUG */

	/*
	 * initial my variables
	 */
	GotSUD = FALSE;
	GotCOD = FALSE;
	Sleep = FALSE;
	systick_raw = FALSE;

	/*
	 * initial LED and debug I/O
	 */
	configure_GPIO();
	MON_LED_ON;

	/*
	 * initial peripherals
	 */
	init_SysTick(sysCLK);
	usart0_init(57600, mainCLK);
	initial_usb_clk_n_pins();
	iap_fup_init();
	pdown_init();
	uhw_Init();			/* initial about USB */
	ucore_InitialCore();

	TD_Init();


	/*
	 * here we go
	 */
	__enable_irq();			/* permit interrupts */
	uhw_Connect(TRUE);		/* connect USB */

	/*
	 * wait for events for endless
	 */
	while(1) {
		/*
		 * we shall sleep when USB-Suspend detected
		 */
		if(Sleep) {
			if(TD_Suspend()==DONE) {
				Sleep = FALSE;
				ucore_TryPowerDown();
			}
			TD_Resume();
		}

		/*
		 * wait for SETUP data available
		 */
		if(GotSUD) {
			GotSUD = FALSE;			/* Clear SUDAV flag */
SET_DEBUG_PORT;
			ucore_SetupCommand();
CLR_DEBUG_PORT;
		}

		/*
		 * wait for Control-OUT data
		 */
		if(GotCOD) {
			GotCOD = FALSE;
SET_DEBUG_PORT;
			ucore_ProceedSetupCommand();
CLR_DEBUG_PORT;
		}

		/*
		 * call user polling routines
		 */
		TD_Poll();
	}

	/* never return from main() */
	while(1)
		;
	//return 0;		/* dummy */
}

/* =============================================================== */

/*
 * configure system clock
 *   (use_PLL != 0)  start SYSOSC and PLL for mainCLK = 48MHz
 *   (use_PLL == 0)  run with IRC 12MHz
 *       ... in both case, sysCLK divider will be set
 */
u_long initial_sysCLK(int use_PLL, u_long divider)
{
	VOLATILE u_long ttt;
	u_long mainCLK;
	u_long sysCLK;

		/*
		 * expect these conditions by cold start
		 */
		//LPC_SYSCTL->SYSAHBCLKDIV = 1;		/* 1/1 (reset value) */
		//LPC_SYSCTL->SYSPLLCLKSEL = 0;		/* sel IRC (reset value) */
		//LPC_SYSCTL->SYSPLLCLKUEN = 0;
		//LPC_SYSCTL->SYSPLLCLKUEN = 1;
		//LPC_SYSCTL->MAINCLKSEL = 0x0;		/* run with IRC (reset value) */
		//LPC_SYSCTL->MAINCLKUEN = 0;
		//LPC_SYSCTL->MAINCLKUEN = 1;
		//LPC_SYSCTL->USBCLKSEL = 0;		/* USB PLL out (reset value) */
		//LPC_SYSCTL->CLKOUTUEN = 0;
		//LPC_SYSCTL->CLKOUTUEN = 1;
		//LPC_SYSCTL->CLKOUTSEL = 0;				/* IRC (reset value) */
		//LPC_SYSCTL->CLKOUTUEN = 0;
		//LPC_SYSCTL->CLKOUTUEN = 1;
		//LPC_FMC->FLASHTIM = xx | FLASHTIM_50MHZ_CPU;	/* (reset value) */

	/*
	 * divider value check
	 */
	if(divider == 0) {
		divider = 1;
	} else if(divider > 255) {
		divider = 255;
	}

	if(use_PLL) {
		/*
		 * start system oscillator with 12MHz X'tal
		 */
		LPC_SYSCTL->PDRUNCFG |= SYSCTL_POWERDOWN_SYSOSC_PD;		/* confirm power off */
		LPC_SYSCTL->SYSOSCCTRL = 0x00;							/* for 12MHz */
		LPC_SYSCTL->PDRUNCFG &= ~SYSCTL_POWERDOWN_SYSOSC_PD;	/* power on SYSOSC */
			/* wait for a while (4x 25msec, by IRC) */
		for(ttt=0; ttt<(LOOP_WAIT_1msec*25); ttt++) {
			;
		}

		/*
		 * connect system oscillator output to SYS PLL
		 */
		LPC_SYSCTL->SYSPLLCLKSEL = 0x01;	/* select SYSOSC */
		LPC_SYSCTL->SYSPLLCLKUEN = 0x00;		/* feed it */
		LPC_SYSCTL->SYSPLLCLKUEN = 0x01;

		/*                              M   Fclkin   FCCO         P
		 * start PLL for mainCLK=48MHz (4 * 12MHz = 192MHz / (2 * 2))
		 */
		LPC_SYSCTL->SYSPLLCTRL = (/*PSEL=*/ 0x01 <<5) | (/*MSEL=*/ 4 -1);
		LPC_SYSCTL->PDRUNCFG &= ~SYSCTL_POWERDOWN_SYSPLL_PD;	/* enable PLL */
			/* wait for PLL lock */
		while(! (LPC_SYSCTL->SYSPLLSTAT & 1) ) {
			;
		}

		/*
		 * it's ready
		 */
		mainCLK = 48*1000*1000;
	}
	else {
		/*
		 * run with IRC 12MHz
		 */
		mainCLK = 12*1000*1000;
	}

	/*
	 * setup divider before using high frequency
	 */
	LPC_SYSCTL->SYSAHBCLKDIV = divider;
	sysCLK = mainCLK / divider;

	if(use_PLL) {
		/*
		 * connect PLL output
		 */
		LPC_SYSCTL->MAINCLKSEL = 0x3;		/* select PLL-output */
		LPC_SYSCTL->MAINCLKUEN = 0;
		LPC_SYSCTL->MAINCLKUEN = 1;
	}

	/*
	 * tune wait for FLASH access
	 */
	ttt = LPC_FMC->FLASHTIM & ~0x3;

	if(sysCLK <= 20*1000*1000) {
		ttt |= FLASHTIM_20MHZ_CPU;
	}
	else if(sysCLK <= 40*1000*1000) {
		ttt |= FLASHTIM_40MHZ_CPU;
	}
	else {		/*(sysCLK <= 50*1000*1000)*/
		ttt |= FLASHTIM_50MHZ_CPU;
	}
	LPC_FMC->FLASHTIM = ttt;

	return sysCLK;
}

/* =============================================================== */

/*
 *
 */
void control_USB_PLL(int start)
{
//	/*
//	 * select USB PLL clock source
//	 *   !! both IRC and sysOSC must be running to switch !!
//	 */
//	LPC_SYSCTL->USBPLLCLKSEL = 1;	/* select sysOSC */
//	LPC_SYSCTL->USBPLLCLKUEN = 0;
//	LPC_SYSCTL->USBPLLCLKUEN = 1;		/* feed it */
// ...... do not touch, use IRC

	if(start) {
		/*                          M   Fclkin   FCCO         P
		 * start USB PLL for 48MHz (4 * 12MHz = 192MHz / (2 * 2))
		 */
		LPC_SYSCTL->USBPLLCTRL = (/*PSEL=*/ 0x01 <<5) | (/*MSEL=*/ 4 -1);
		LPC_SYSCTL->PDRUNCFG &= ~SYSCTL_POWERDOWN_USBPLL_PD;	/* enable PLL */
				/* wait for PLL lock */
		while(! (LPC_SYSCTL->USBPLLSTAT & 1) ) {
			;
		}
	}
	else {
		/*
		 * stop it
		 */
		LPC_SYSCTL->PDRUNCFG |= SYSCTL_POWERDOWN_USBPLL_PD;		/* diable PLL */
	}
}

/* =============================================================== */

/*
 *
 */
void configure_GPIO(void)
{
//	/*
//	 * prevent the the open-drain pins from floating, if needed
//	 *   (PIO0_4 and PIO0_5)
//	 */
//#define OPEN_DRAIN_PINS   ((1<<4) | (1<<5))
//	LPC_GPIO->CLR[0]  = OPEN_DRAIN_PINS;	/* both Low */
//	LPC_GPIO->DIR[0] |= OPEN_DRAIN_PINS;	/* start to drive */

	/*
	 * setup GPIO first (connect to pins by default, with pull-up-R)
	 */
	LPC_GPIO->DIR[0] &= ~GPI_pSW;		/* confirm input */

	LPC_GPIO->CLR[0] = GPO_LED;			/* initial to LOW */
	LPC_GPIO->DIR[0] |= GPO_LED;		/* start to drive */

	LPC_GPIO->CLR[0] = GPO_DEBUG;		/* initial to LOW */
	LPC_GPIO->DIR[0] |= GPO_DEBUG;		/* start to drive */

	/*
	 * confirm I/O configuration and choose pull-R mode
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

	LPC_IOCON->PIO0[1] = 0x0090;	/* GPI_pSW, with pull-up-R */
	LPC_IOCON->PIO0[7] = 0x0080;	/* GPO_LED,   without pull-R */
	LPC_IOCON->PIO0[9] = 0x0080;	/* GPO_DEBUG, without pull-R */

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);
}

/* =============================================================== */

/*
 * Initialize pin and clocks for USB port
 */
void initial_usb_clk_n_pins(void)
{
	VOLATILE u_long ttt;

	/* enable USB main clock */
		// (mainCLK is already selected)

	/* Enable AHB clock to the USB block and USB RAM. */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_USB);
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_USBRAM);
	//LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_RAM1);	/* unused */

	/* power UP USB Phy */
	LPC_SYSCTL->PDRUNCFG &= ~SYSCTL_POWERDOWN_USBPAD_PD;
		/* wait a little, to make sure */
	for(ttt=0; ttt<(LOOP_WAIT_1msec*10); ttt++) {
		;
	}

	/*
	 * config other pins
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

	LPC_IOCON->PIO0[3] = 0x0089;	/* USB_VBUS with pull-down-R */
	LPC_IOCON->PIO0[6] = 0x0081;	/* USB_CONNECT without pull-R */

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);
}

/* =============================================================== */

/*
 * run SysTick timer
 */
void init_SysTick(u_long sysCLK)
{
	/*
	 * setup SysTick timer
	 */
	SysTick->LOAD = sysCLK / SYSTICK_Hz / 2 - 1;	/* use half-clock */
	SysTick->VAL = 0;			/* clear the counter (value is ignored) */
		/*
		 * 30,000,000 = 0x01C9_C380  >  (1<<24)
		 * 16,777,215 = 0x00FF_FFFF  =  (1<<24)
		 * 15,000,000 = 0x00E4_E1C0  <  (1<<24) ... half-clock covers 1Hz
		 */
	/*
	 * setup NVIC
	 */
	NVIC_SetPriority(SysTick_IRQn, PRI_SYSTICK);

	/*
	 * enable intr and run it
	 */
	SysTick->CTRL = 0x03;		/* run it with half-clock, intr enabled */
}

/*
 * interrupt service routine
 */
void SysTick_Handler(void)
{
	systick_raw = TRUE;
		/*
		 * interrupt request flag may be cleared automatically
		 */
}

/* =============================================================== */

/*
 * try to catch exceptions
 */

void def_Exception_Handler(void)
{
	RPRINTF("! some Exception\r\n");
	while(1)
		;		/* halt at here */
}

void def_NMI_Handler(void)
{
	RPRINTF("! NMI\r\n");
	while(1)
		;		/* halt at here */
}

void def_HardFault_Handler(void)
{
	RPRINTF("! HardFault\r\n");
	while(1)
		;		/* halt at here */
}

void def_SVCall_Handler(void)
{
	RPRINTF("! SVCall\r\n");
	while(1)
		;		/* halt at here */
}

void def_PendSV_Handler(void)
{
	RPRINTF("! PendSV\r\n");
	while(1)
		;		/* halt at here */
}

/* =============================================================== */
