/*
 * pdown.c - control power-down
 *   by uratan! 2020.7.1
 */
#include "common.h"
#include <chip.h>

	/*
	 * use 'Power-down mode' as low-power mode
	 */

/*
 * my local debug flags
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf
#define PUTCHAR    if(myUDEBUG)cm_putchar
	/*
	 * !CAUTION! this macro makes wrong 'else' like: "if(x) RPRINTF() else ..."
	 */

/*
 * use PIN_INT0 for wakeup
 */
#define PIN_INT0    (1<<0)

/*
 * assign PIOn_x to PIN_INT0
 */
#if DEVELOP_by_SOLO
# define WAKEUP_PIO    1		/* PIO0_1 (ISP-en for test) */
#else /* DEVELOP_by_SOLO */
#define WAKEUP_PIO    18		/* PIO0_18 (RXD) */
#endif /* DEVELOP_by_SOLO */

/*
 * functions
 */
void pdown_init(void);
void PIN_INT0_IRQ_Handler(void);
int power_down(int en_pint);


/* =============================================================== */

/*
 *
 */
void pdown_init(void)
{
	/*
	 * initial Pin interrupts registers
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_PINT);

	LPC_PININT->ISEL = 0;		/* edge sensitive */
	LPC_PININT->IENR = 0;		/* disable rising edge */
	LPC_PININT->IENF = 0;		/* disable falling edge */
	LPC_PININT->RISE = 0xff;	/* clear rising edge detection */
	LPC_PININT->FALL = 0xff;	/* clear falling edge detection */
	LPC_PININT->IST = 0xff;		/* clear both edge detection */
		/*
		 * pin function  GPIO ɂȂĂȂĂoł܂
		 */

	/*
	 * config PIN_INT0 for fall-edge of PIO0_x
	 */
	LPC_SYSCTL->PINTSEL[0] = WAKEUP_PIO;	/* select PIO0_x for PIN_INT0 */
	LPC_PININT->IST = PIN_INT0;			/* clear both request */
	LPC_PININT->IENF = PIN_INT0;		/* enable falling edge */

	NVIC_SetPriority(PIN_INT0_IRQn, PRI_PIN_INT0);
	NVIC_DisableIRQ(PIN_INT0_IRQn);
	LPC_SYSCTL->STARTERP0 = PIN_INT0;			/* wakeup by pin intr 0 */

	/*
	 * config about USB_WAKEUP_IRQ (about USB need_clock wire)
	 */
	/*
	 * initial about USB need_clock signal
	 */
	LPC_SYSCTL->USBCLKCTRL = 0x02;			/* POL_CLK=1, AP_CLK=0 */

	NVIC_SetPriority(USB_WAKEUP_IRQn, PRI_USB_WAKEUP);
	NVIC_DisableIRQ(USB_WAKEUP_IRQn);
	LPC_SYSCTL->STARTERP1 = SYSCTL_WAKEUP_USB_WAKEUP;	/* by both USB_WAKEUP_IRQ and USB0_IRQ */

	return;
}

/*
 * handler is prepared, but I will control not to use this
 */
void PIN_INT0_IRQ_Handler(void)
{
RPRINTF("P-%d\r\n", __get_IPSR());
	LPC_PININT->IST = PIN_INT0;		/* need to clear my request */
}

/*
 * enter 'power-down mode'
 */
	/*
	 * To allow the LPC11U3x/2x/1x to wake up from Deep-sleep or Power-down
	 * mode on USB activity, complete the following steps:
	 *
	 *   1. Set bit AP_CLK in the USBCLKCTRL register (Table 41) to 0 (default)
	 *      to enable automatic control of the USB need_clock signal.
	 *   2. Wait until USB activity is suspended by polling the DSUS bit
	 *      in the DEVCMDSTAT register (DSUS = 1).
	 *   3. The USB need_clock signal will be deasserted after another 2 ms.
	 *      Poll the USBCLKST register until the USB need_clock status bit
	 *      is 0 (Table 42).
	 *   4. Once the USBCLKST register returns 0, enable the USB activity
	 *      wake-up interrupt in the NVIC (# 30) and clear it.
	 *   5. Set bit POL_CLK in the USBCLKCTRL register to 1 to trigger the
	 *      USB activity wake-up interrupt on the rising edge of the USB
	 *      need_clock signal.
	 *   6. Enable the wake-up from Deep-sleep or Power-down modes on this
	 *      interrupt by enabling the USB need_clock signal in the STARTERP1
	 *      register (Table 44, bit 19).
	 *
	 *   7. Enter Deep-sleep or Power-down modes by writing to the PCON
	 *      register.
	 *   8. Execute a WFI instruction.
	 */
int power_down(int en_pint)
{
	u_long imf;
	uint32_t backup_of_SysTick_CTRL;
	uint32_t backup_of_SSPnCLKDIV;
	int wakeup_by_pint;
	VOLATILE u_long ttt;

	/*
	 * mask all interrupts by core
	 */
	imf = IntDisable();

	/*
	 * stop SysTick because it can wake me up from low-power mode
	 */
	backup_of_SysTick_CTRL = SysTick->CTRL;
	SysTick->CTRL = 0x00;		/* stop systick & disable intr */
			/*
			 * have some wait until above memory write will take effect
			 */
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
			asm("nop");
	/*
	 * then clear SysTick pending request by SCB->ICSR (if exist)
	 */
	SCB->ICSR = SCB_ICSR_PENDSTCLR_Msk;

	/*
	 * set to turn off BOD and WWDT while Power-down
	 */
	LPC_SYSCTL->PDSLEEPCFG = 0xffff;

	/*
	 * confirm LOCK bit of WWDT
	 */
//	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_WDT);
//	LPC_WWDT->MOD = 0;
//	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_WDT);
//////    this is done in iap_fup_init()

	/*
	 * configure analog power switchs after wake-up
	 */
	LPC_SYSCTL->PDWAKECFG = LPC_SYSCTL->PDRUNCFG;	/* recover current state */

	/*
	 * disable all unrelated interrups because for reason detection
	 */
		// SysTick is already stopped
	NVIC_DisableIRQ(UART0_IRQn);
	NVIC_DisableIRQ(USB0_IRQn);		/* USB_WAKEUP@STARTERP1 accepts this also */

	/*
	 * wait for USB need_clock signal deasserted,
	 * then activate USB_WAKEUP_IRQ (POL_CLK is already initialied to 1)
	 */
	for(ttt=0; ttt<(LOOP_WAIT_1msec*100); ttt++) {
		if(! (LPC_SYSCTL->USBCLKST & 1) ) {
			break;
		}
	}
	NVIC_ClearPendingIRQ(USB_WAKEUP_IRQn);
	NVIC_EnableIRQ(USB_WAKEUP_IRQn);

	/*
	 * enable pin-int if specified
	 */
	if(en_pint) {
		LPC_PININT->IST = PIN_INT0;					/* clear request */
		NVIC_ClearPendingIRQ(PIN_INT0_IRQn);		/* clear pending request */
		NVIC_EnableIRQ(PIN_INT0_IRQn);				/* enable it */
	}

	/*
	 * switch mainCLK back for IRC  (48MHz --> 12MHz)
	 */
	LPC_SYSCTL->MAINCLKSEL = 0x0;		/* run with IRC */
	LPC_SYSCTL->MAINCLKUEN = 0;
	LPC_SYSCTL->MAINCLKUEN = 1;
		/*
		 * adjust SSP xfer clock to keep DEBUG-uart speed
		 */
		backup_of_SSPnCLKDIV = LPC_SYSCTL->SSPnCLKDIV;
		LPC_SYSCTL->SSPnCLKDIV /= 4;		/* max 1/255 */
			/*
			 * untouch about ROM-wait
			 */
			//LPC_FMC->FLASHTIM = ...;

	/*
	 * select Power-down mode for WFI instruction and
	 * set SLEEPDEEP bit in the ARM Cortex-M0 SCR register
	 */
	LPC_PMU->PCON = PMU_PCON_PM_POWERDOWN  | PMU_PCON_SLEEPFLAG;
	SCB->SCR = SCB_SCR_SLEEPDEEP_Msk;

RPRINTF("now Power-down\r\n%%");		/* clock stops at last '%' */

			/*
			 * now I can Power-down
			 */
			__WFI();

SET_DEBUG_PORT;
//RPRINTF("woke-up\r\n");

	/*
	 * recover to select normal sleep for WFI instruction
	 */
	SCB->SCR = 0x00;
	LPC_PMU->PCON = PMU_PCON_PM_SLEEP;

	/*
	 * get wake up reason
	 */
	wakeup_by_pint = (SCB->ICSR >> SCB_ICSR_VECTPENDING_Pos) & 0x3f;
				/* give same priority for USB_WAKEUP_IRQn / PIN_INT0_IRQn */

	/*
	 * recover clocks if once power-down
	 */
	if(LPC_PMU->PCON & PMU_PCON_SLEEPFLAG) {
//RPRINTF(" slept\r\n");
		/*
		 * had entered Power-down mode once
		 */

		/*
		 * wait for SYSOSC/SYSPLL stable
		 */
				/*
				 * 1msec ΕA悤
				 * Ƃ肠 IRC 12MHz  4msec ҂Ƃ
				 * (wake-up ṽVAʐM͊ۂƃXg鎞)
				 */
		for(ttt=0; ttt<LOOP_WAIT_1msec; ttt++) {
			;
		}
			/* wait for PLL lock */
		while(! (LPC_SYSCTL->SYSPLLSTAT & 1) ) {
			;
		}
	}
	else {
//RPRINTF(" wake-up kept\r\n");
		/*
		 * low-power-mode was skipped because
		 * wake-up request was already exist
		 */
	}
CLR_DEBUG_PORT;

		/*
		 * recover SSPn xfer clock
		 */
		LPC_SYSCTL->SSPnCLKDIV = backup_of_SSPnCLKDIV;

	/*
	 * recover mainCLK (12MHz --> 48MHz)
	 */
	LPC_SYSCTL->MAINCLKSEL = 0x3;		/* select PLL-output */
	LPC_SYSCTL->MAINCLKUEN = 0;
	LPC_SYSCTL->MAINCLKUEN = 1;

	/*
	 * recover interrupt condition
	 */
	if(en_pint) {
		NVIC_DisableIRQ(PIN_INT0_IRQn);		/* no need while normal jobs */
			/*
			 * handler is not called when Disabled
			 *   (regardless of pending-request)
			 */
	}
	NVIC_DisableIRQ(USB_WAKEUP_IRQn);
	NVIC_EnableIRQ(USB0_IRQn);
	NVIC_EnableIRQ(UART0_IRQn);

	/*
	 * recover SysTick
	 */
	SysTick->CTRL = backup_of_SysTick_CTRL & 0x07;

	/*
	 * recover interrupt mask state
	 */
	IntRestore(imf);

RPRINTF("by %d\r\n", wakeup_by_pint);
	wakeup_by_pint = (wakeup_by_pint == 16) ? 1 : 0;		/* 1 if PIN_INT0 */

	return(wakeup_by_pint);
}

/* =============================================================== */
