/*
 * periph.c : UZK main processes
 *   by uratan! 2020.9.18
 */
#include "common.h"
#include <chip.h>
#include "dscr.h.S"
#include "usb.h"
#include "usbjmptb.h"

/*
 * my local debug flags
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf
	/*
	 * !CAUTION! this macro makes wrong 'else' like: "if(x) RPRINTF() else ..."
	 */

/*
 * make error situation for debug
 */
#define ERR_DEBUG  0
#if ERR_DEBUG
int x4 = 0;		// DataOUT @ set_report (5) ... NG
int x5 = 0;		// StatusIN @ set_report (3) ... OK
#endif /* ERR_DEBUG */

/*
 * class request for HID device
 */
#define UR_GET_REPORT			0x01
#define UR_SET_REPORT			0x09
#define UR_GET_IDLE				0x02
#define UR_SET_IDLE				0x0a
#define     IDLE_TIME_RATIO        4	/* [msec/1] */
#define UR_GET_PROTOCOL			0x03
#define UR_SET_PROTOCOL			0x0b

/*
 * functions
 */
/* in myself */
void TD_Init(void);
void TD_Poll(void);
void poll_serial(void);
void poll_timer0(void);
void poll_hid_trigger(void);
int TD_Suspend(void);
void TD_Resume(void);
int DR_SetupCommand(void);
void DR_ProceedSetupCommand(void);
int DR_GetDescriptor(void);
void xbzero(u_char *dst, u_long len);
void xmemcpy(u_char *dst, u_char *src, u_long len);
void isr_Reset_Event(u_long unused);
void isr_Suspend_Event(u_long unused);
void isr_Resume_Event(u_long unused);
void isr_EP1_in(u_long unused);
void isr_EP2_in(u_long unused);
void isr_EP3_in(u_long unused);
void isr_EP4_in(u_long unused);
/* jump tables */
const INTR_subFUNCs USB_Jump_Table[USB_N_JT];
const INTR_subFUNCs USB_Jump_Table_EP[USB_N_EP];

/*
 * my variables
 */
/*
 * interface/endpoint status
 */
u_char Protocol[N_IFs];		/* (nothing is influenced by this value) */
u_char Idle_Rate[N_IFs];	/* (but has no idle timer function) */
u_char work_INxBUF[N_IFs][MAX_IN_DATA_LEN];
u_char INxBUF_backup[N_IFs][MAX_IN_DATA_LEN];
u_char work_buff[MAX_DATA_LEN];

#define HT_NO_TRIGGER          0
#define HT_WORK_INxBUF_READY   1	/* working INxBUF is ready to send */
//#define HT_IDLE_TIME_ELAPSED   2	/* idle time elapsed, enable tansfer */
u_char HIDx_trigger[N_IFs];

/*
 * list of endpoint addresses
 */
const u_char EP_IN[N_IFs] = {
	DSCR_ADDR_EP1i,	/* KB */
	DSCR_ADDR_EP2i,	/* CTRL */
	DSCR_ADDR_EP3i,	/* MS */
#if A_MOUSE
	DSCR_ADDR_EP4i,	/* A_MS */
#endif /* A_MOUSE */
};

/*
 * consts for accessing Nth endpoint data length
 */
const u_char HIDx_in_data_len[N_IFs] = {
	HID1_IN_DATA_LEN,
	HID2_IN_DATA_LEN,
	HID3_IN_DATA_LEN,
#if A_MOUSE
	HID4_IN_DATA_LEN,
#endif /* A_MOUSE */
};

/*
 * serial comminucation protocol
 *                                        +----+
 *  reset                                 |    | ... data sent by SET_REPORT
 *    ++++++                              +----+
 *    |0xff|                              ++++++
 *    ++++++                              |    | ... added for communication
 *                                        ++++++
 *    send LED status (payload is 2 byte)
 *       ++++++  ++++++ +----+
 *       |'L' |  |'L' | |    |
 *       ++++++  ++++++ +----+
 *    send keycodes (payload is 8 bytes)
 *       +----+  +----+----+----+----+----+----+----+----+
 *       |'K' |  |    |    |    |    |    |    |    |    |
 *       +----+  +----+----+----+----+----+----+----+----+
 *    send mouse action (payload is 5 bytes)
 *       +----+  +----+----+----+----+----+
 *       |'M' |  |    |    |    |    |    |
 *       +----+  +----+----+----+----+----+
 *    send ABS mouse action (payload is 7 bytes)
 *       +----+  +----+----+----+----+----+----+----+
 *       |'A' |  |    |    |    :    |    :    |    |
 *       +----+  +----+----+----+----+----+----+----+
 *    send control status (payload is 2 byte)
 *       ++++++  +----+ +----+
 *       |'C' |  |'C' | |    |
 *       ++++++  +----+ +----+
 *               |<------------------------------------->|
 *              send by text encoded / passed thru each Endpoints
 *                (ascii-hex with lower case)
 */
	/*
	 * L[{[hE}EXƂĔF HID1/HID3 ͓Rwb_[Ȃ
	 * Reĉ݂肵܂Ap HID2 ʂƂ
	 * wb_[łB܂A
	 * 'K', 'M', 'A', 'C' ͐e PC  HID2 ɏނƂ 9oCgŒ蒷
	 * ł serial ɗƂɏk߂Ă܂B
	 */
			/*
			 * ==> A_MOUSE f[^Ƃ 0x00-0xff ϓɈ̂ŁA
			 *     thumbUZK ł payload f[^ hex-text ő邱ƂƂ
			 *     fW^AC\[^ŒʐM[g 6{ 57600bps 
			 *        ( wake-up ԍۂɖ肪ł邩)
			 *        (  Ẑ 57600bps   6oCg(1msec) )
			 *        (  ̂ 57600bps  18oCg(3msec) )
			 */
#if DEVELOP_by_SOLO
# define HEAD_RESET  '?'		/* for test */
#else /* DEVELOP_by_SOLO */
#define HEAD_RESET  0xff	/* you can not pass this as contents !!! */
#endif /* DEVELOP_by_SOLO */
#define HEAD_LEDS   'L'
#define HEAD_CTRL   'C'		/* misc control for future use */
#define HEAD_KEYS   'K'
#define HEAD_MOUS   'M'
#define HEAD_A_MS   'A'

/*
 * serial1 comminucation receiver states
 */
#define RECV_for_HID1     IF_HID1_KB		/* shall be 0 */
#define RECV_for_HID2     IF_HID2_CTRL		/* shall be 1 */
#define RECV_for_HID3     IF_HID3_MS		/* shall be 2 */
#if A_MOUSE
#define RECV_for_HID4     IF_HID4_A_MS		/* shall be 3 */
#endif /* A_MOUSE */
#define RECV_NULL_or_IDLE   8	/* idle state */
#define RECV_RESET          9	/* header gotten */
u_char recv_target;				/* I have to handle HEAD_A_MS when !A_MOUSE */
u_char recv_ptr;
u_char recv_ptr_eo;				/* even/odd */
u_char recv_prev_bin;			/* first upper half bin value */

/*
 * for blinking monitor LED
 */
u_char mon_led_timer;

/*
 * for proceeding SET_REPORT request
 */
BOOL cont_is_hid1;
u_char cont_req_len;

/*
 * externs
 */
/* in usart0.c */
extern int u0_peek_recv(void);
extern void u0_set_to_send(u_char c);
extern void u0_set_to_send_hex(u_char c);
/* in usbhw.c */
extern u_long uhw_read_SE_stat(u_char EPaddr);
extern void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len);
extern void uhw_arm_WriteEP(u_char EPaddr);
/* in usbcore.c */
extern void ucore_start_DataIN(u_char *buff, u_long len);
extern int ucore_start_DataOUT(u_char *buff, u_long len);
extern void ucore_start_StatusIN(void);
extern void ucore_STALL_EP0(void);
/* in iap_fup.c */
extern int fup_erase(void);
extern int fup_write(u_char *buff);
extern void fup_update_master(void);

/* in dscr.S */
extern u_long *HIDDscr1, *EpI1Dscr;
extern u_long *ReportDscr1, *ReportDscr1Len;
extern u_long *HIDDscr2, *EpI2Dscr;
extern u_long *ReportDscr2, *ReportDscr2Len;
extern u_long *HIDDscr3, *EpI3Dscr;
extern u_long *ReportDscr3, *ReportDscr3Len;
#if A_MOUSE
extern u_long *HIDDscr4, *EpI4Dscr;
extern u_long *ReportDscr4, *ReportDscr4Len;
#endif /* A_MOUSE */
/* in main.c */
extern VOLATILE int systick_raw;
/* in usbcore.c */
extern u_char **USB_list_of_EpDscrs;
extern u_char USB_Configuration;	/* 0 or configuration number */
extern u_char SETUPDAT[8];

/*
 * list of endpoint descriptors
 */
u_char *my_endpoints[N_IFs +1] = {
	(u_char *)&EpI1Dscr,
	(u_char *)&EpI2Dscr,
	(u_char *)&EpI3Dscr,
#if A_MOUSE
	(u_char *)&EpI4Dscr,
#endif /* A_MOUSE */
	NULL,
};

/* =============================================================== */
/* =============================================================== */

/*
 * Called once at startup
 */
void TD_Init(void)
{
	int i;

	/*
	 * USB endpoints are initialized according to Endpoint-Descriptor
	 * when SET_CONFIG request will come.
	 */
	USB_list_of_EpDscrs = &my_endpoints[0];

	/*
	 * initial each interface
	 */
	for(i=0; i<N_IFs; i++) {
		xbzero(INxBUF_backup[i], HIDx_in_data_len[i]);

		Protocol[i] = 1;		/* boot device */

		HIDx_trigger[i] = HT_NO_TRIGGER;
	}

	/*
	 * HID spec sheet says that
	 *   recommanded default idle rate is
	 *     500ms for keyboard, 0ms for mouse.
	 */
	Idle_Rate[IF_HID1_KB]   = 500/*[msec]*/ / IDLE_TIME_RATIO;
	Idle_Rate[IF_HID2_CTRL] =   0/*[msec]*/ / IDLE_TIME_RATIO;
	Idle_Rate[IF_HID3_MS]   =   0/*[msec]*/ / IDLE_TIME_RATIO;
#if A_MOUSE
	Idle_Rate[IF_HID4_A_MS] =   0/*[msec]*/ / IDLE_TIME_RATIO;
#endif /* A_MOUSE */

	/*
	 * initial others
	 */
	recv_target = RECV_NULL_or_IDLE;		/* reset state */

	mon_led_timer = 0;

	/*
	 * initial my peripherals
	 */
	// (done all in main())

	return;
}

/*
 * Called repeatedly while the device is idle
 */
void TD_Poll(void)
{
	/*
	 * when idle, this function takes:
	 *  17.6usec (24MHz+24MHz), 7.8usec (60MHz+30MHz), where 60usec by EZ-USB
	 */
	/*
	 * is something coming from another side of UZK ?
	 */
	poll_serial();

	/*
	 * process events about time
	 */
	poll_timer0();

	/*
	 * check data-ready of HID endpoint
	 */
	poll_hid_trigger();

	return;
}

/*
 * check event arival from other UZK
 */
void poll_serial(void)
{
	int d;

	/*
	 * is something coming from another side of UZK ?
	 */
	d = u0_peek_recv();
	if(d == -1) {
		return;
	}

	/*
	 * something gotten
	 */
//PUTCHAR(d);
	MON_LED_OFF;	/* turn off LED cause something act */

	switch(recv_target) {
	case RECV_NULL_or_IDLE:
			if(d == HEAD_RESET) {
				recv_target = RECV_RESET;
			} else {
				;							/* do not change state */
			}
			break;

	case RECV_RESET:
			switch(d) {
			case HEAD_RESET:
					;							/* keep this state */
					break;
			case HEAD_KEYS:
					recv_target = RECV_for_HID1;	/* change state */
					break;
			case HEAD_LEDS:
			case HEAD_CTRL:
					recv_target = RECV_for_HID2;	/* change state */
					break;
			case HEAD_MOUS:
					recv_target = RECV_for_HID3;	/* change state */
					break;
#if A_MOUSE
			case HEAD_A_MS:
					recv_target = RECV_for_HID4;	/* change state */
					break;
#endif /* A_MOUSE */
			default:
					recv_target = RECV_NULL_or_IDLE;	/* reset state */
					break;
			}

			/* this initialze is for RECV_for_HIDx state */
			recv_ptr = 0;
			recv_ptr_eo = 0;
			break;

#if A_MOUSE
	case RECV_for_HID4:
#endif /* A_MOUSE */
	case RECV_for_HID3:
	case RECV_for_HID2:
	case RECV_for_HID1:
			if(d == HEAD_RESET) {
				recv_target = RECV_RESET;
				break;
			}

			if(! is_hex(d)) {			/* pass lower-case only */
				recv_target = RECV_NULL_or_IDLE;	/* reset state */
				break;
			}

			d = to_bin(d);
			if(recv_ptr_eo == 0) {
				recv_prev_bin = d << 4;		/* save upper half */
				recv_ptr_eo = 1;
			}
			else {
				d = recv_prev_bin | d;		/* bind both */
				recv_ptr_eo = 0;

//RPRINTF("-> [%d] %02x\r\n", recv_ptr, d);

				/*
				 * threw previous work_INxBUF value away
				 * wheather they are still valid ot not.
				 */
				HIDx_trigger[recv_target] &= ~HT_WORK_INxBUF_READY;
					/*
					 * work_INxBUF Ƀf[^܂ĂĂ
					 * ̂ĂĂ܂܂B
					 */

				work_INxBUF[recv_target][recv_ptr++] = d;
				if(recv_ptr == HIDx_in_data_len[recv_target]) {
					HIDx_trigger[recv_target] |= HT_WORK_INxBUF_READY;
					recv_target = RECV_NULL_or_IDLE;	/* reset state */
				}
			}
			break;

	default:
			recv_target = RECV_NULL_or_IDLE;	/* reset state */
			break;
	}
}

/*
 * generate events by timer
 */
void poll_timer0(void)
{
	/*
	 * measure report interval time
	 */
	if(systick_raw) {
		/*
		 * clear the flag
		 */
		systick_raw = FALSE;

		/*
		 * turn on monitor every interval time
		 */
		if(++mon_led_timer == 25/*[msec]*/) {
			mon_led_timer = 0;	/* reset timer */
			MON_LED_ON;			/* turn on LED here */
		}
	}
}

/*
 * send HID usages if misc condition met
 */
void poll_hid_trigger(void)
{
	int i;

	/*
	 * loop for all endpoints
	 */
	for(i=0; i<N_IFs; i++) {
		/*
		 * arm the endpoint if condition met
		 */
		if((USB_Configuration == 0) || (uhw_read_SE_stat(EP_IN[i]) & (1<<31)))
		{
			/*
			 * the endpoint is already armed, but IN request does not
			 * come yet. do NOT touch INxBUF because I don't know
			 * when IN request will come.
			 */
			;	/* do nothing now, keep state and try in next TD_Poll() loop */
					/*
					 * zXgf[^ǂ݂ɗ邩Ȃ
					 *  INxBUF ̓eύXłȂBύXĂ
					 * r̃f[^ǂ܂邩Ȃ
					 */
		} else {
			/*
			 * when new data arrived, copy contents from work_INxBUF.
			 *    (HT_WORK_INxBUF_READY is prior to HT_IDLE_TIME_ELAPSED)
			 */
			if(HIDx_trigger[i] & HT_WORK_INxBUF_READY) {
				/*
				 * save data for GET_REPORT request
				 * (keep only button state, because others are *differential*)
				 */
					/*
					 * Ԍo߂قɂđ́AWf[^() 0 Ƃ
					 * (L[{[h̏ꍇɂ́uOƓL[vɂȂ̂ŖȂ)
					 * (A}EX̏ꍇ́uOƓړʁvɂȂĂ܂)
					 */
				xmemcpy(INxBUF_backup[i], work_INxBUF[i], HIDx_in_data_len[i]);
				if(i==IF_HID3_MS) {
					INxBUF_backup[IF_HID3_MS][1] = 0;	/* X */
					INxBUF_backup[IF_HID3_MS][2] = 0;	/* Y */
					INxBUF_backup[IF_HID3_MS][3] = 0;	/* WHEEL */
					INxBUF_backup[IF_HID3_MS][4] = 0;	/* wheel-tile */
				}
#if A_MOUSE
				else if(i==IF_HID4_A_MS) {
					INxBUF_backup[IF_HID4_A_MS][1] = 0;	/* WHEEL */
					INxBUF_backup[IF_HID4_A_MS][6] = 0;	/* wheel-tile */
				}
#endif /* A_MOUSE */

				/*
				 * set the data to send
				 */
				uhw_WriteEP(EP_IN[i], work_INxBUF[i], HIDx_in_data_len[i]);
			}
//			else if(HIDx_trigger[i] & HT_IDLE_TIME_ELAPSED) {
//				uhw_WriteEP(EP_IN[i], INxBUF_backup[i], HIDx_in_data_len[i]);
//			}

			/*
			 * arm the endpoint
			 */
			if(HIDx_trigger[i]) {
				HIDx_trigger[i] = HT_NO_TRIGGER;

				uhw_arm_WriteEP(EP_IN[i]);
			}
		}
	}
}

/*
 * Called before the device goes into suspend mode
 */
int TD_Suspend(void)
{
	MON_LED_OFF;	/* turn off LED */

	return(DONE);
}

/*
 * Called after the device resumes
 */
void TD_Resume(void)
{
	recv_target = RECV_NULL_or_IDLE;		/* reset state */
	return;
}

/* =============================================================== */
/*
 * HID Requset Support Requirements
 *
 *                     Get      Set      Get      Set      Get      Set
 *   Device Type       Report   Report   Idle     Idle     Protocol Protocol
 * -------------------------------------------------------------------------
 *  Boot Mouse         Required Optional Optional Optional Required Required
 *  Non-Boot Mouse     Required Optional Optional Optional Optional Optional
 *  Boot Keyboard      Required Optional Required Required Required Required
 *  Non-Boot Keyboard  Required Optional Required Required Optional Optional
 *  Other Device       Required Optional Optional Optional Optional Optional
 */
	/*
	 * ܂ HID ̎dl悭ǂłꂽ܂B
	 * ϐ Protocol[] ̓e͂͂Ă܂
	 * œς͂Ă܂B
	 */
/*
 * Device Request hook function
 */
int DR_SetupCommand(void)
{
	u_char target_hid;
	u_char ans_len, req_len, i;
	BOOL is_hid1;

	MON_LED_OFF;	/* turn off LED cause something act */

	/*
	 * feed class request only
	 */
	if((SETUPDAT[0] & UT_TYPE_MASK) != UT_CLASS)
		return(YET);

	/*
	 * which interface is target ?
	 */
	if((SETUPDAT[5] != 0x00) || (SETUPDAT[4] >= N_IFs))
		goto exit_with_stall;

	target_hid = SETUPDAT[4];

	/*
	 * I shall not handle so long data
	 */
	if(SETUPDAT[7] != 0x00)
		goto exit_with_stall;

	req_len = SETUPDAT[6];
	ans_len = 0;			/* no need to answer */

	switch(SETUPDAT[1]) {
	case UR_GET_REPORT:			/* 0x01 mandatory */
RPRINTF("GET_REPO\r\n");
			ans_len = HIDx_in_data_len[target_hid];
			xmemcpy(work_buff, INxBUF_backup[target_hid], ans_len);
			break;

	case UR_GET_IDLE:			/* 0x02 optional */
RPRINTF("GET_IDLE\r\n");
			work_buff[0] = Idle_Rate[target_hid];
			ans_len = 1;
			break;

	case UR_SET_IDLE:			/* 0x0a optional */
RPRINTF("SET_IDLE - %02x\r\n", SETUPDAT[3]);
			/*
			 * ignore if shoter than Polling interval
			 * in endpoint descriptor (10msec)
			 */
			if(SETUPDAT[3] == 0)
				Idle_Rate[target_hid] = 0;	/* indefinite */
			else if(SETUPDAT[3] < 3)
				Idle_Rate[target_hid] = 3;	/* minimum */
			else
				Idle_Rate[target_hid] = SETUPDAT[3];

			break;

	case UR_GET_PROTOCOL:		/* 0x03 mandatory if boot dev */
RPRINTF("GET_PROT\r\n");
			work_buff[0] = Protocol[target_hid];
			ans_len = 1;
			break;

	case UR_SET_PROTOCOL:		/* 0x0b mandatory if boot dev */
RPRINTF("SET_PROT - %02x\r\n", SETUPDAT[2]);
			Protocol[target_hid] = SETUPDAT[2];
			break;

	case UR_SET_REPORT:			/* 0x09 optional */
RPRINTF("SET_REPO -");
			if(target_hid == IF_HID1_KB) {
				is_hid1 = TRUE;
				i = HID1_OUT_DATA_LEN;
			} else if(target_hid == IF_HID2_CTRL) {
				is_hid1 = FALSE;
				i = HID2_OUT_DATA_LEN;
			} else {
				goto exit_with_stall;
			}

//			if(req_len != i)
//				goto exit_with_stall;

			/*
			 * arm OUT endpoint, then wait
			 */
#if ERR_DEBUG
	if(++x4 == 5) {
		x4 = 0;
	//rprintf("do\r\n");
		goto exit_with_stall;
	}
#endif /* ERR_DEBUG */
			if(ucore_start_DataOUT(work_buff, i))
				goto exit_with_stall;

			cont_is_hid1 = is_hid1;
			cont_req_len = req_len;
			goto continue_after_data_out;

	default:
			goto exit_with_stall;
	}

	if(ans_len != 0) {
		ucore_start_DataIN(work_buff, ans_len);
	} else {
		ucore_start_StatusIN();
	}

continue_after_data_out:
	return(DONE);

	/*
	 * ---------------------------------
	 *    entry for return with error
	 * ---------------------------------
	 */
exit_with_stall:

	ucore_STALL_EP0();
	return(DONE);
}

/*
 * process end-half of SETUP comamnd which has DataOUT
 */
void DR_ProceedSetupCommand(void)
{
	BOOL is_hid1;
	u_char req_len, i, jj;

	is_hid1 = cont_is_hid1;
	req_len = cont_req_len;

	if(!is_hid1) {
		/*
		 * add these command for firmare update
		 */
		switch(work_buff[0]) {
		case 'e':
RPRINTF("e");
				fup_erase();
				ucore_start_StatusIN();
				return;
		case 'w':
RPRINTF("w");
				fup_write(&work_buff[1]);
				ucore_start_StatusIN();
				return;
		case 'r':
RPRINTF("r\r\n");
				ucore_start_StatusIN();
				fup_update_master();
				return;
		}
	}

#if myUDEBUG
for(i=0; i<req_len; i++)
	rprintf(" %02x", work_buff[i]);
rprintf("\r\n");
#endif /* myUDEBUG */

	/*
	 * rewind communication every time
	 * cause I don't know partner's state.
	 * he may NOT be in `RECV_NULL_or_IDLE`
	 * by power-down or suspend of my side.
	 */
		/*
		 * serial ̌ɂ鑊ǂȏԂɂ邩
		 * Ȃ̂ŁAMƂɏ߂𑗂邱Ƃɂ
		 */
if(DEVELOP_by_SOLO) {
	u0_set_to_send('\r');
	u0_set_to_send('\n');
}
	u0_set_to_send(HEAD_RESET);
		/*
		 * because here is NOT interrupt layer, I can wait.
		 * so u0_set_to_send() waits when send buff is full.
		 * at the result, transaction with host shall be late.
		 */
			/*
			 * ͂ Setup Command ̏́A荞ݏł
			 * Ȃ|[ÖꕔȂ̂ŁAԂĂ
			 * ̂łBȂ̂łłɑMobt@ς
			 *  u0_set_to_send() ֐̓obt@󂭂܂ő҂悤
			 * ܂BN҂̂ƂƁASET_REPORT 
			 * s PC ҂܂B
			 */

	if(is_hid1) {
		/*
		 * KB LED control is naked 1byte
		 */
		u0_set_to_send(HEAD_LEDS);		/* this is for serial comm */
		u0_set_to_send_hex(HEAD_LEDS);	/* this is for return data */
		jj = 0;
		req_len = HID1_OUT_DATA_LEN;
	} else {
		/*
		 * data from host have serial comminication header
		 * HEAD_KEYS/HEAD_MOUS/HEAD_CTRL/HEAD_A_MS on the top
		 */
		switch(work_buff[0]) {
		case HEAD_KEYS:
				u0_set_to_send(work_buff[0]);
				jj = 1;
				req_len = HID1_IN_DATA_LEN;		/* KB */
				break;

		case HEAD_CTRL:
				u0_set_to_send(work_buff[0]);	/* this is for serial comm */
				jj = 0;							/* this is for return data */
				req_len = HID2_IN_DATA_LEN;		/* CTRL */
				break;

		case HEAD_MOUS:
				u0_set_to_send(work_buff[0]);
				jj = 1;
				req_len = HID3_IN_DATA_LEN;		/* MS */
				break;
//#if A_MOUSE							/* pass this if I am !A_MOUSE */
		case HEAD_A_MS:
				u0_set_to_send(work_buff[0]);
				jj = 1;
				req_len = HID4_IN_DATA_LEN;		/* A_MS */
				break;
//#endif /* A_MOUSE */
		default:
				goto exit_with_stall;
		}
	}

	/*
	 * send payload contents by ascii encoded
	 */
	for(i=0; i<req_len; i++) {
		u0_set_to_send_hex(work_buff[jj++]);
	}

#if ERR_DEBUG
	if(++x5 == 3) {
		x5 = 0;
		goto exit_with_stall;
	}
#endif /* ERR_DEBUG */
	ucore_start_StatusIN();
	return;

exit_with_stall:
	ucore_STALL_EP0();
	return;
}

/*
 * Get Descriptor Device Request hook function
 */
int DR_GetDescriptor(void)
{
	u_char *dscr;
	u_long len;

	if((SETUPDAT[5] != 0x00) || (SETUPDAT[4] >= N_IFs)) {
		return(YET);
	}

	if(SETUPDAT[3]==UDESC_HID/*0x21*/) {
RPRINTF("UDESC_HID\r\n");

		switch(SETUPDAT[4]) {
		case IF_HID1_KB:
				dscr = (u_char *)&HIDDscr1;
				len = dscr[/*bLength*/ 0];
				break;
		case IF_HID2_CTRL:
				dscr = (u_char *)&HIDDscr2;
				len = dscr[/*bLength*/ 0];
				break;
		case IF_HID3_MS:
				dscr = (u_char *)&HIDDscr3;
				len = dscr[/*bLength*/ 0];
				break;
#if A_MOUSE
		case IF_HID4_A_MS:
				dscr = (u_char *)&HIDDscr4;
				len = dscr[/*bLength*/ 0];
				break;
#endif /* A_MOUSE */
		}

		ucore_start_DataIN(dscr, len);
		return(DONE);
	}

	if(SETUPDAT[3]==UDESC_REPORT/*0x22*/) {
RPRINTF("UDESC_REPO\r\n");

		switch(SETUPDAT[4]) {
		case IF_HID1_KB:
				dscr = (u_char *)&ReportDscr1;
				len = (u_long)&ReportDscr1Len;
				break;
		case IF_HID2_CTRL:
				dscr = (u_char *)&ReportDscr2;
				len = (u_long)&ReportDscr2Len;
				break;
		case IF_HID3_MS:
				dscr = (u_char *)&ReportDscr3;
				len = (u_long)&ReportDscr3Len;
				break;
#if A_MOUSE
		case IF_HID4_A_MS:
				dscr = (u_char *)&ReportDscr4;
				len = (u_long)&ReportDscr4Len;
				break;
#endif /* A_MOUSE */
		}

		ucore_start_DataIN(dscr, len);
		return(DONE);

	}

	return(YET);		// if(SETUPDAT[3]==UDESC_PHYSICAL/*0x23*/) etc.
}

/* =============================================================== */

/*
 * function to clear memories
 */
void xbzero(u_char *dst, u_long len)
{
	while(len--) {
		*dst = 0x00;
		dst++;
	}

	return;
}

/*
 * function to copy memories
 */
void xmemcpy(u_char *dst, u_char *src, u_long len)
{
	while(len--) {
		*dst = *src;
		dst++;
		src++;
	}

	return;
}

/* =============================================================== */
/*
 * list of USB device events callbacks
 */
const INTR_subFUNCs USB_Jump_Table[USB_N_JT] = {
/*0:USB_JT_RESET  */	isr_Reset_Event,
/*1:USB_JT_SUSPEND*/	isr_Suspend_Event,
/*2:USB_JT_RESUME */	isr_Resume_Event,
/*3:USB_JT_POWER  */	NULL  /*isr_Power_Event*/,
/*4:USB_JT_SOF    */	NULL  /*isr_SOF_Event*/,
};

/*
 * USB Reset Event Callback
 */
void isr_Reset_Event(u_long unused)
{
RPRINTF("!Res\r\n");
}

/*
 * USB Suspend Event Callback
 */
void isr_Suspend_Event(u_long unused)
{
RPRINTF("!Sus\r\n");
}

/*
 * USB Resume Event Callback
 */
void isr_Resume_Event(u_long unused)
{
RPRINTF("!Rsm\r\n");
}

/*
 * use these if needed
 */
//void isr_Power_Event(u_long vbus_on) {;}
//void isr_SOF_Event(u_long frame_num) {;}

/* =============================================================== */
/*
 * list of USB Endpoint events callbacks
 */
const INTR_subFUNCs USB_Jump_Table_EP[USB_N_EP] = {
/*00: 0out*/	NULL,		/* control out endpoint, unused */
/*80: 0in*/		NULL,		/* control in endpoint, unused */
/*01: 1out*/	 NULL,		/* interrupt OUT */
/*81: 1in*/		isr_EP1_in,	/* interrupt IN */	/*KB*/
/*02: 2out*/	 NULL,		/* interrupt OUT */
/*82: 2in*/		isr_EP2_in,	/* interrupt IN */	/*CTRL*/
/*03: 3out*/	 NULL,		/* interrupt OUT */
/*83: 3in*/		isr_EP3_in,	/* interrupt IN */	/*MS*/
/*04: 4out*/	 NULL,		/* interrupt OUT */
/*84: 4in*/		isr_EP4_in,	/* interrupt IN */	/*A_MS*/
};

/*
 *
 */
void isr_EP1_in(u_long unused)
{
RPRINTF("!1i\r\n");
	/*
	 * this entry is called after IN request has success...
	 */
		/*
		 *  IN 荞݂́AIN NGXgăobt@
		 * ɂȂƂ(̃f[^pӂ邽߂)Ăяo܂B
		 */
}

/*
 *
 */
void isr_EP2_in(u_long unused)
{
RPRINTF("!2i\r\n");
	/*
	 * this entry is called after IN request has success...
	 */
}

/*
 *
 */
void isr_EP3_in(u_long unused)
{
RPRINTF("!3i\r\n");
	/*
	 * this entry is called after IN request has success...
	 */
}

/*
 *
 */
void isr_EP4_in(u_long unused)
{
RPRINTF("!4i\r\n");
	/*
	 * this entry is called after IN request has success...
	 */
}

/* =============================================================== */
