/*
 * usart0.c - use usart0 to communicate with the other side
 *   by uratan! 2020.6.24
 */
#include "common.h"
#include <chip.h>

/*
 * my local debug flags
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf
#define PUTCHAR    if(myUDEBUG)cm_putchar
	/*
	 * !CAUTION! this macro makes wrong 'else' like: "if(x) RPRINTF() else ..."
	 */

/*
 * choise FIFO depth
 */
#define N_BUFF          64		/* must be 2^x */
#define BUFF_PTR_MASK  (N_BUFF-1)
		/*
		 * at close, the application may write to all 3 devices at once.
		 *   '?' 'K' 0000000000000000 ... 18chars
		 *   '?' 'M' 0000000000       ... 12chars
		 *   '?' 'A' 00000000000000   ... 16chars ... 46chars in total
		 * the FIFO depth should be greater than this
		 */
/*
 * functions
 */
void usart0_init(u_long baudrate, u_long mainCLK);
int u0_peek_recv(void);
void u0_set_to_send(u_char c);
void u0_set_to_send_hex(u_char c);
void USART_IRQ_Handler(void);

/*
 * valiables
 */
static u_char send_buff[N_BUFF];
static u_long send_in_ptr, send_out_ptr;
static VOLATILE u_long n_send;		/* need 'VOLATILE' because polling this */
static int ti1_will_come;

static u_char recv_buff[N_BUFF];
static u_long recv_in_ptr, recv_out_ptr;
static u_long n_recv;

/* =============================================================== */

/*
 * initial USART
 */
void usart0_init(u_long baudrate, u_long mainCLK)
{
	u_long Fdiv;

	/*
	 * enable clock for USART
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_UART0);

	/*
	 * enable clock for baudrate generator
	 */
	LPC_SYSCTL->USARTCLKDIV = 1;		/* mainCLK(12MHz or 48MHz) thru */

	/*
	 * setup usart
	 */
	LPC_USART->IER = 0x00;		/* kill interrupts */
	LPC_USART->FDR = 0x85;		/* use Fractional Divider */
	LPC_USART->LCR = 0x80;		/* DLAB=1 */
	Fdiv = 8 * mainCLK / 16 / baudrate / 13;	/* calc division from baudrate */
	LPC_USART->DLM = Fdiv / 256;
	LPC_USART->DLL = Fdiv % 256;
	LPC_USART->LCR = 0x03;		/* DLAB=0, 8bits, no-parity, 1stopbit */
	//LPC_USART->FCR = 0x07;		/* use FIFO, clear buffers */
	LPC_USART->FCR = 0x00;		/* do not use FIFO */
		/*
		 * EFCR=0x01 ł RX FIFOATX FIFO ɗL
		 * EFCR=0x00 ł RX FIFO ͖ɂ邪 TX FIFO ͗L̂܂
		 * EgƂāA
		 *   + THRE ҂ł THR+TSR  2oCgobt@Ȃ
		 *   + TEMT ҂ł     TSR  1oCgobt@Ȃ
		 * ETX FIFO full tO͂ȂAꂽ
		 *   u THRE  16oCg܂ŏĂ悢v
		 *     (x[X̏ꍇATX FIFO 󂢂͕KvȂ)
		 *     (x[X̏ꍇATHRE ŃCxg)
		 */
	/*
	 * initial both FIFO
	 */
	send_in_ptr = send_out_ptr = 0;
	n_send = 0;
	ti1_will_come = FALSE;		/* no THRE intr will come */

	recv_in_ptr = recv_out_ptr = 0;
	n_recv = 0;

	/*
	 * connect USART to the pins
	 */
	LPC_SYSCTL->SYSAHBCLKCTRL |= (1 << SYSCTL_CLOCK_IOCON);

	LPC_IOCON->PIO0[18] = 0x0091;		/* select RxD with PullUP */
	LPC_IOCON->PIO0[19] = 0x0081;		/* select TxD */

	LPC_SYSCTL->SYSAHBCLKCTRL &= ~(1 << SYSCTL_CLOCK_IOCON);

	/*
	 * at last, enable interrupts
	 */
	LPC_USART->IER = UART_IER_RBRINT | UART_IER_THREINT;	/* enable interrupts */

	/*
	 * setup NVIC to enable interrupt
	 */
	NVIC_SetPriority(UART0_IRQn, PRI_UART0);
	NVIC_EnableIRQ(UART0_IRQn);

	return;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * check receive buffer
 */
int u0_peek_recv(void)
{
	u_long imf;
	u_char c;

	/*
	 * chech buffer empty
	 */
	if(n_recv == 0)
		return(-1);		/* no data has come */

	/*
	 * take 1char out from buffer
	 */
	imf = IntDisable();				/* disable interrupts */

	c = recv_buff[recv_out_ptr++];
	n_recv--;
	recv_out_ptr &= BUFF_PTR_MASK;

	IntRestore(imf);				/* (may) enable interrupts */

	return(c);
}

/*
 * pile the data on send buffer
 */
void u0_set_to_send(u_char c)
{
	u_long imf;

	/*
	 * wait if there is no space in send buffer,
	 * i.e, keep host waiting on process of SET_REPORT request
	 */
	while(n_send == N_BUFF)
		;
				/*
				 * may freeze here when call me from interrupt context
				 * and if USART_IRQ_Handler() can not interrupt here.
				 */
	/*
	 * try to set the char
	 */
	imf = IntDisable();				/* disable interrupts */

	if(ti1_will_come) {
		/*
		 * I am doing transfer now, so THRE intr will flush the buffer
		 */
		send_buff[send_in_ptr++] = c;
		n_send++;
		send_in_ptr &= BUFF_PTR_MASK;
PUTCHAR('s');
	} else {
		/*
		 * transmitter is idle, no THRE intr will come, so
		 */
		LPC_USART->THR = c;			/* send directly here */
		ti1_will_come = TRUE;
PUTCHAR('d');
	}

	IntRestore(imf);				/* (may) enable interrupts */

	return;
}

/*
 * pile the data on send buffer by hex-text
 */
void u0_set_to_send_hex(u_char c)
{
	u_char d;

	/* upper half */
	d = (c >> 4) & 0x0f;
	if(d < 10) {
		d += '0';
	} else {
		d += 'a' - 10;		/* use lower case */
	}
	u0_set_to_send(d);

	/* lower half */
	d = (c >> 0) & 0x0f;
	if(d < 10) {
		d += '0';
	} else {
		d += 'a' - 10;		/* use lower case */
	}
	u0_set_to_send(d);
}


/*
 * interrupt service routine
 */
void USART_IRQ_Handler(void)
{
	u_char iid, c;

	/*
	 * take out interrupt reason
	 */
	iid = LPC_USART->IIR & UART_IIR_INTID_MASK;		/* Interrupt Identification */

	switch(iid) {
	case UART_IIR_INTID_RDA:
	case UART_IIR_INTID_CTI:
			/*
			 * in case of Receive Data Available, or
			 *            Character Time-out indicator
			 */
PUTCHAR('R');
			/*
			 * read the data, then put it on receive buffer
			 */
			c = LPC_USART->RBR;		/* <== this clears interrupt request */

			if(n_recv == N_BUFF) {
				/*
				 * buffer overflow, only ignore
				 */
PUTCHAR('O');
			} else {
				/*
				 * save received data on buffer
				 */
				recv_buff[recv_in_ptr++] = c;
				n_recv++;
				recv_in_ptr &= BUFF_PTR_MASK;
			}
			break;

	case UART_IIR_INTID_THRE:
			/*
			 * in case of THRE, transmit holding register empty
			 *   the request is automatically cleared when reading U1IIR above
			 */
PUTCHAR('T');
			/*
			 * confirm status in the LSR
			 */
			if(LPC_USART->LSR & UART_LSR_THRE) {
				if(n_send == 0) {
					/*
					 * flushing buffer is done, THRE intr chain is stopped here
					 */
					ti1_will_come = FALSE;
				} else {
					/*
					 * send data on buffer
					 *   this will generate next THRE intr
					 */
					LPC_USART->THR = send_buff[send_out_ptr++];
					n_send--;
					send_out_ptr &= BUFF_PTR_MASK;
				}
			}
			break;

	default:
			/*
			 * something is wrong !!
			 */
PUTCHAR('X');
			while(1)
				;
	}
}

/* =============================================================== */
