/*
 * usbhw.c : LPC11U35 USB logic driver
 *   by uratan! 2020.9.22
 */
#include "common.h"
#include <chip.h>
#include "dscr.h.S"
#include "usbjmptb.h"
#include <string.h>		// memcpy()

/*
 * my local debug flag
 */
#define myUDEBUG  (0 && gUDEBUG)
#define RPRINTF    if(myUDEBUG)rprintf
#define PUTCHAR    if(myUDEBUG)cm_putchar

/*
 * register bitmaps
 */
	/* for DEVCMDSTAT */
#define DCS_DEV_ADDR_MASK  (0x7f)	/* 0 RW  USB device address */
#define DCS_DEV_EN        (1<< 7)	/* 0 RW  USB device enable */
#define DCS_SETUP         (1<< 8)	/* 0 RWC SETUP token received */
#define DCS_PLL_ON        (1<< 9)	/* 0 RW  Always PLL Clock on */
#define DCS_LPM_SUP       (1<<11)	/* 1 RW  LPM Supported */
#define DCS_INTONNAK_AO   (1<<12)	/* 0 RW  Interrupt on NAK for intr and bulk OUT EP */
#define DCS_INTONNAK_AI   (1<<13)	/* 0 RW  Interrupt on NAK for intr and bulk IN EP */
#define DCS_INTONNAK_CO   (1<<14)	/* 0 RW  Interrupt on NAK for control OUT EP */
#define DCS_INTONNAK_CI   (1<<15)	/* 0 RW  Interrupt on NAK for control IN EP */
#define DCS_DCON          (1<<16)	/* 0 RW  Device status - connect */
#define DCS_DSUS          (1<<17)	/* 0 RW  Device status - suspend */
#define DCS_LPM_SUS       (1<<19)	/* 0 RW  Device status - LPM Suspend */
#define DCS_LPM_REWP      (1<<20)	/* 0 RO  LPM Remote Wake-up Enabled by host */
#define DCS_DCON_C        (1<<24)	/* 0 RWC Device status - connect change */
#define DCS_DSUS_C        (1<<25)	/* 0 RWC Device status - suspend change */
#define DCS_DRES_C        (1<<26)	/* 0 RWC Device status - reset change */
#define DCS_VBUSDEBOUNCED (1<<28)	/* 0 RO  Vbus is detected or not */

#define DCS_1SAFEbits  (DCS_DSUS)
#define DCS_0SAFEbits  (DCS_DRES_C | DCS_DSUS_C | DCS_DCON_C   | DCS_SETUP)

	/* mainly for EPINUSE, EPBUFCFG, INTSTAT, INTEN, INTSETSTAT, INTROUTING, EPTOGGLE */
#define bitEP0OUT   (1<<0)
#define bitEP0IN    (1<<1)
#define bitEP1OUT   (1<<2)
#define bitEP1IN    (1<<3)
#define bitEP2OUT   (1<<4)
#define bitEP2IN    (1<<5)
#define bitEP3OUT   (1<<6)
#define bitEP3IN    (1<<7)
#define bitEP4OUT   (1<<8)
#define bitEP4IN    (1<<9)
#define bits_EPall ((1<<10)-1)		/* 0x3ff, EP0_OUT -- EP4_IN */
#define bitFRM_INT (1<<30)
#define bitDEV_INT (1<<31)

	/* for Endpoint Command/Status list */
#define BM_LIST      (0x100-1)		/* boundary mask for ep-list */
#define BM_BUFF       (0x40-1)		/* boundary mask for ep buffer */

#define ECS_ACTIVE     (1<<31)		/* Activate the endpoint / Armed */
#define ECS_DISABLED   (1<<30)		/* out of use, no response */
#define ECS_enabled      (0)		/* (answer NAK) */
#define ECS_STALL      (1<<29)		/* Stall */
#define ECS_TogRESET   (1<<28)		/* Toggle Reset */
#define ECS_TogVALUE   (1<<27)		/* Toggle value */
#define ECS_RateFB   ECS_TogVALUE	/* Rate Feedback mode */
#define ECS_TypeISO    (1<<26)		/* Isochronous endpoint */
#define ECS_CSALL_MASK (0x3f<<26)	/* all bits */
#define ECS_pos_NBytes     16		/* 0x3ff */
#define ECS_NBytes_MASK  0x3ff		/* 0x3ff */
#define ECS_pos_Offset      0		/* 0xffff */
#define ECS_Offset_MASK 0xffff		/* 0xffff */
	/*
	 * endpoint list command combination under my driver
	 *
	 *           | A | D | S | TR | TV | T | N Bytes | Addr Offset |
	 * ---------------------------------------------------------------------
	 * DISABLED:   0   1   0    0    0   0     0        0    (no response)
	 * enabled:    0   0   0    0    0   0     0        0    (return NAK)
	 * ARMED(IN):  1   0   0    r    0   0     n        a    (respond IN)
	 * ARMED(OUT): 1   0   0    r    0   0     m        b    (respond OUT)
	 * STALL:      0   0   1    0    0   0     0        0    (return STALL)
	 *
	 *   ... hardware  host ̗v󂯂 Y list ǂ
	 *       ߂AƍlB(v driven)
	 *   ... N Bytes  Addr Offset  TR  Arm/ACTIVATE ƂȊO͕sv
	 *   ... n  Mf[^Am ͎󂯓\
	 *             (n,m  dec  a,b  inc )
	 *   ... ACTIVE  hardware Asoftwre  list ̏֎~
	 *             (USB RESET ͗OƂc)
	 *             ( host ̗vȂΏĂȂ͂)
	 *   ... DISABLED ł ACTIVATE ł邪A͎gȂ
	 *   ... A  S 𗼕Ă A D悳 M A  S ݂̂
	 *       c邪 ͎gȂ
	 */

/*
 * functions
 */
int uhw_Init(void);
void uhw_Connect(int conn);
void uhw_WakeUp(void);
void uhw_SetAddress(u_char addr);
void uhw_Configure(int cfg);
void uhw_EnableEP(u_char EPaddr, u_long max_packet_size);
u_long uhw_IsEnabledEP(u_char EPaddr);
u_long uhw_read_SE_stat(u_char EPaddr);
u_long uhw_IsStallEP(u_char EPaddr);
void uhw_SetStallEP(u_char EPaddr);
void uhw_ClrStallEP(u_char EPaddr);
void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len);
void uhw_arm_WriteEP(u_char EPaddr);
u_long uhw_ReadEP(u_char EPaddr, u_char *pData, int capacity);
void uhw_ReadEP_finish(u_char EPaddr);
#if WA_X_STALL_EP0_OUT
void uhw_arm_EP0_OUT_insteadof_STALL(void);
#endif /* WA_X_STALL_EP0_OUT */
void uhw_ReadSETUP_finish(void);
static u_long get_EPnum(u_long EPaddr);
void USB_IRQ_Handler(void);

	/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * 'EPaddr' is USB spec style Endpoint address.
	 *   Endpoint number is at [bit3..0], bit 7 has direction
	 * 'EPnum' is LPC2148 physical Endpoint number.
	 *   Endpoint number is at [bit4..1], bit 0 has direction
	 *vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv */

/*
 *   (boundary: 0x0100)
 */
#define USB_STACK_MEM_BASE      0x20004000			/* at USB SRAM */
#define USB_STACK_MEM_SIZE      0x0800				/* all of USB SRAM */

/*
 * misc parameter table referenced by epnum
 */
struct _ep_list {
	u_long buff_addr;		/* buffer location */
	u_long buff_size;		/* buffer size */
VOLATILE u_long *list_at;	/* endpoint list location */
	u_long tr_ofs;			/* have buff offset and next TR bit */
	u_long len;				/* send/receive length */
} uhw_eplist[USB_N_EP];

/*
 * externs
 */
/* in usbcore.c */
extern void ucore_ResetCore(void);
extern void ucore_SuspendCore(void);
extern void ucore_isr_EP0_in(u_long unused);
extern void ucore_isr_EP0_out(u_long unused);
/* in periph.c */
extern const INTR_subFUNCs USB_Jump_Table[USB_N_JT];
extern const INTR_subFUNCs USB_Jump_Table_EP[USB_N_EP];

/*
 * variables
 */
/* in main.c */
extern VOLATILE int GotSUD;		/* SetUp Data */
extern VOLATILE int GotCOD;		/* Control-OUT Data */
extern VOLATILE int Sleep;		/* shall Sleep becaue in Suspend mode */


/* =============================================================== */

/*
 * initialize USB hardware
 */
int uhw_Init(void)
{
	u_long memL, memB;
	VOLATILE u_long *list_at;
	int i;
	struct _ep_list *epl;

	/*
	 * initial USB device controller
	 */
	LPC_USB->DEVCMDSTAT = 0x00000000;	/* confirm stop to initial */

	//LPC_USB->EPTOGGLE					/*  (read-only) */
	LPC_USB->INTROUTING = 0x00000000;	/* route all to IRQ */
	//LPC_USB->INTSETSTAT				/*  (needless) */
	LPC_USB->INTEN      = 0x00000000;	/* disable all */
	LPC_USB->INTSTAT    = 0xc00003ff;	/* clear all request */
	LPC_USB->EPBUFCFG   = 0x00000000;	/* use as single buffer */
	LPC_USB->EPINUSE    = 0x00000000;	/*     all buffer0 */
	LPC_USB->EPSKIP     = 0x00000000;	/* (reset value) */
	//LPC_USB->LPM						/* Link Power Management register */
	//LPC_USB->INFO						/* USB Info register */

	/*
	 * initial memories
	 */
	memL = USB_STACK_MEM_BASE;			/* mem for ep-list */
	memL = (memL + BM_LIST) & ~BM_LIST;	/* boundary: 0x100 */
	LPC_USB->EPLISTSTART = memL;		/* use from here */

	list_at = (u_long *)memL;			/* init all command/status memory */
	for(i=0; i<(USB_N_EP * 2); i++) {
		*list_at++ = ECS_DISABLED;
	}

	memB = (u_long)list_at;				/* mem for buffer */
	memB = (memB + BM_BUFF) & ~BM_BUFF;	/* boundary: 0x40 */
	LPC_USB->DATABUFSTART = memB;		/* use this area */

	/*
	 * alloc buffers for all endpoint here
	 */
	list_at = (u_long *)memL;			/* point top of list */
	i = 0;
	epl = &uhw_eplist[i];
	for(   ; i<USB_N_EP; i++) {			/* EP0_OUT, EP0_IN, ..., EP4_IN */
		epl->buff_addr = memB;			/* buffer location */
		epl->buff_size = DSCR_MaxPkSIZE;	/* buffer size */
		memB += epl->buff_size;			/* advance for next */
		memB = (memB + BM_BUFF) & ~BM_BUFF;	/* boundary: 0x40 */
		epl->list_at = list_at;			/* list location */
		list_at += 2;					/* no use second of double buffer */
		epl->tr_ofs = (epl->buff_addr >> 6) & ECS_Offset_MASK;
		epl++;
	}

#if myUDEBUG
if(1) {
	rprintf(" DATABUFSTART: %08x\r\n", LPC_USB->DATABUFSTART);
	rprintf(" EPLISTSTART : %08x\r\n", LPC_USB->EPLISTSTART);
	epl = &uhw_eplist[0];
	for(i=0; i<USB_N_EP; i++) {
		rprintf("%2d: %08x %08x %08x %02x\r\n", i,
			epl->list_at, epl->buff_addr, epl->tr_ofs, epl->buff_size);
		epl++;
	}
}
#endif /* myUDEBUG */

	if(memB > (USB_STACK_MEM_BASE + USB_STACK_MEM_SIZE)) {
		RPRINTF("memory overflow\r\n");
		return(FALSE);		/* memory error */
	}

	/*
	 * initial control pipe and SETUP token
	 */
	list_at = uhw_eplist[0].list_at;
	*list_at++ = ECS_enabled;				/* enable EP0_OUT */
	*list_at++ = uhw_eplist[0].tr_ofs;		/* use same buff for SETUP */
	*list_at++ = ECS_enabled;				/* enable EP0_IN */

	/*
	 * configure about interrupts
	 */
	LPC_USB->INTEN = bitDEV_INT |
					((USB_Jump_Table[USB_JT_SOF] != NULL) ? bitFRM_INT : 0) |
							bitEP0OUT | bitEP0IN;
	NVIC_SetPriority(USB0_IRQn, PRI_USB);
	NVIC_EnableIRQ(USB0_IRQn);

	/*
	 * start operation
	 */
	LPC_USB->DEVCMDSTAT = DCS_DRES_C | DCS_DSUS_C | DCS_DCON_C | DCS_DSUS | DCS_SETUP | DCS_DEV_EN;

	return(TRUE);
}

/*
 * connect USB device to the host
 *   pass 1 to Connect, 0 to Disconnect
 */
void uhw_Connect(int conn)
{
	u_long imf;
	u_long dcs_base;

	imf = IntDisable();

	if(conn) {
		dcs_base = (LPC_USB->DEVCMDSTAT & ~DCS_0SAFEbits) | DCS_1SAFEbits;
		LPC_USB->DEVCMDSTAT = dcs_base | DCS_DCON;
	} else {
		dcs_base = (LPC_USB->DEVCMDSTAT & ~DCS_0SAFEbits) | DCS_1SAFEbits;
		LPC_USB->DEVCMDSTAT = dcs_base & ~DCS_DCON;
	}

	IntRestore(imf);
}

/*
 * issue USB-Remote-Wakeup action to wake up the host
 */
	/*
	 * To issue a remote wake-up when the USB activity is suspended,
	 * complete the following steps:
	 *
	 *   1. Set bit AP_CLK in the USBCLKCTRL register to 0 (Table 41, default)
	 *      to enable automatic control of the USB need_clock signal.
	 *   2. When it is time to issue a remote wake-up, turn on the USB clock
	 *      and enable the USB clock source.
	 *   3. Force the USB clock on by writing a 1 to bit AP_CLK (Table 41,
	 *      bit 0) in the USBCLKCTRL register.
	 *   4. Write a 0 to the DSUS bit in the DEVCMDSTAT register.
	 *   5. Wait until the USB leaves the suspend state by polling the
	 *      DSUS bit in the DEVCMDSTAT register (DSUS = 0).
	 *   6. Clear the AP_CLK bit (Table 41, bit 0) in the USBCLKCTRL to
	 *      enable automatic USB clock control.
	 */
void uhw_WakeUp(void)
{
	u_long imf;
	u_long dcs_base;
	VOLATILE u_long ttt;

	imf = IntDisable();

	/*
	 * force USB need_clock signal HIGH
	 */
	LPC_SYSCTL->USBCLKCTRL |= 0x01;			/* AP_CLK=1 */

	/*
	 * issue remote-wakeup operation
	 */
	dcs_base = (LPC_USB->DEVCMDSTAT & ~DCS_0SAFEbits) | DCS_1SAFEbits;
	LPC_USB->DEVCMDSTAT = dcs_base & ~DCS_DSUS;

	/*
	 * wait for leaving suspend state
	 */
	for(ttt=0; ttt<(LOOP_WAIT_1msec*1000); ttt++) {
		if(! (LPC_USB->DEVCMDSTAT & DCS_DSUS) ) {
			break;
		}
	}

	/*
	 * recover USB need_clock handling
	 */
	LPC_SYSCTL->USBCLKCTRL &= ~0x01;		/* AP_CLK=0 */

	IntRestore(imf);
}

/*
 * set device address
 */
void uhw_SetAddress(u_char addr)
{
	u_long imf;
	u_long dcs_base;

	imf = IntDisable();

	dcs_base = (LPC_USB->DEVCMDSTAT & ~DCS_0SAFEbits) | DCS_1SAFEbits;
	LPC_USB->DEVCMDSTAT = (dcs_base & ~DCS_DEV_ADDR_MASK) | addr | DCS_SETUP;
		/*
		 * software must program the new address before completing the
		 * status phase of the SetAddress Control Request.
		 *     ==> DCS_SETUP=0 is needed to update the address
		 */

	IntRestore(imf);
}

/*
 * Configure the USB device
 *   pass 1 to Configure, 0 to De-configure
 *   (in both case, non-control-Endpoints are disabled)
 */
void uhw_Configure(int cfg)
{
	u_long imf;
	u_long epnum;
	struct _ep_list *epl;

	imf = IntDisable();

	/*
	 * disable all Endpoints from EP1out to EP4in
	 *  (except control-pipe)
	 */
	epnum = 2;
	epl = &uhw_eplist[epnum];
	for(   ; epnum<USB_N_EP; epnum++) {
		LPC_USB->INTEN &= ~(1 << epnum);

		if(*epl->list_at & ECS_ACTIVE) {
			LPC_USB->EPSKIP = (1 << epnum);			/* de-activate */
			while(LPC_USB->EPSKIP & (1 << epnum)) {
				;
			}
			LPC_USB->INTSTAT = (1 << epnum);		/* clear request */
		}

		*epl->list_at = ECS_DISABLED;
		//epl->tr_ofs |= ECS_TogRESET;			/* do this at enable timing */
	}

	IntRestore(imf);
}

/*
 * realize to enable the Endpoint
 *   (nor armed)
 */
void uhw_EnableEP(u_char EPaddr, u_long max_packet_size)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	imf = IntDisable();

	if(*epl->list_at & ECS_DISABLED) {
		LPC_USB->INTSTAT = (1 << epnum);		/* clear request */
		LPC_USB->INTEN  |= (1 << epnum);
		*epl->list_at = ECS_enabled;
		if(epnum >= 2) {
			epl->tr_ofs |= ECS_TogRESET;	/* reset toggle-bit at next xfer */
		}
	} else {
		;		/* already enabled */
	}

	IntRestore(imf);
}

/*
 * check the Endpoint is enabled or not
 *   return:  1 if enabled
 */
u_long uhw_IsEnabledEP(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long enabled;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	enabled = (*epl->list_at & ECS_DISABLED) ? FALSE : TRUE;

	return(enabled);
}

/*
 * read Endpoint status
 */
u_long uhw_read_SE_stat(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long val;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	val = *epl->list_at;					/* return command/status */

	return(val);
}

/*
 * check STALL bit of the Endpoint
 *   return:  1 if stalled
 */
u_long uhw_IsStallEP(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long val;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	val = (*epl->list_at & ECS_STALL) ? TRUE : FALSE;	/* check STALL bit */

	return(val);
}

	/*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	 * 'ST' bit is at: (1<<1) in Select-Endpoint-status value.
	 *                 (1<<0) in Set-Endpoint-status register.
	 *vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv */

/*
 * set STALL bit of the Endpoint
 */
void uhw_SetStallEP(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	imf = IntDisable();

	if(*epl->list_at & ECS_ACTIVE) {
		LPC_USB->EPSKIP = (1 << epnum);		/* de-activate */
		while(LPC_USB->EPSKIP & (1 << epnum)) {
			;
		}
		LPC_USB->INTSTAT = (1 << epnum);		/* clear request */
			/*
			 * might clear pending request by
			 *   NVIC_ClearPendingIRQ(USB0_IRQn);
			 */
	}

	if(*epl->list_at & (ECS_DISABLED | ECS_STALL)) {
		;							/* not enabled / already STALLed */
	} else {
		*epl->list_at = ECS_STALL;			/* set STALL bit */
	}

	IntRestore(imf);
}

/*
 * clear STALL bit of the Endpoint
 */
void uhw_ClrStallEP(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

		/*
		 * CLEAR_FEATURE Endpoint Halt shall do:
		 *   Set STALL bit (S) to 0.
		 *   Set toggle reset bit (TR) to 1 and set toggle value bit (TV) to 0.
		 */

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	imf = IntDisable();

	if(*epl->list_at & ECS_ACTIVE) {
		LPC_USB->EPSKIP = (1 << epnum);		/* de-activate */
		while(LPC_USB->EPSKIP & (1 << epnum)) {
			;
		}
		LPC_USB->INTSTAT = (1 << epnum);		/* clear request */
	}

	if(*epl->list_at & ECS_DISABLED) {
		;									/* not enabled */
	} else if(*epl->list_at & ECS_STALL) {
		*epl->list_at = ECS_enabled;		/* clear STALL bit */
	}

	if(epnum >= 2) {
		epl->tr_ofs |= ECS_TogRESET;	/* reset toggle-bit at next xfer */
	}

	IntRestore(imf);
}

/*
 * write data to IN Endpoint buffer
 */
void uhw_WriteEP(u_char EPaddr, u_char *pData, u_long len)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	if(len > epl->buff_size) {
		len = epl->buff_size;
	}

	imf = IntDisable();
	memcpy((u_char *)epl->buff_addr, pData, len);
	epl->len = len;
	IntRestore(imf);
}

/*
 * permit hardware buffer to send the data
 */
void uhw_arm_WriteEP(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	if(*epl->list_at & (ECS_ACTIVE | ECS_DISABLED | ECS_STALL)) {
		;				/* ignore */
RPRINTF("xxx");
	} else {
		imf = IntDisable();
		*epl->list_at = ECS_ACTIVE | epl->tr_ofs | (epl->len << ECS_pos_NBytes);
		epl->tr_ofs &= ~ECS_TogRESET;
		IntRestore(imf);
	}
}

/*
 * read data from OUT Endpoint buffer
 *   pass your buffer capacity to avoid buffer-overflow
 *   return: number of bytes received
 */
u_long uhw_ReadEP(u_char EPaddr, u_char *pData, int capacity)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long residue, len;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	if((epnum == 0) && (LPC_USB->DEVCMDSTAT & DCS_SETUP)) {
		len = 8;						/* fixed 8byte */
	} else {
		residue = (*epl->list_at >> ECS_pos_NBytes) & ECS_NBytes_MASK;
		len = epl->len - residue;
	}
	if(len > capacity) {
		len = capacity;
	}

	imf = IntDisable();
	memcpy(pData, (u_char *)epl->buff_addr, len);
	IntRestore(imf);

	return(len);
}

/*
 * permit hardware buffer to receive next OUT
 */
void uhw_ReadEP_finish(u_char EPaddr)
{
	u_long epnum;
	struct _ep_list *epl;
	u_long imf;

	epnum = get_EPnum(EPaddr);
	epl = &uhw_eplist[epnum];

	if(*epl->list_at & (ECS_ACTIVE | ECS_DISABLED | ECS_STALL)) {
		;				/* ignore */
RPRINTF("xx");
	} else {
		imf = IntDisable();
		epl->len = epl->buff_size;
		*epl->list_at = ECS_ACTIVE | epl->tr_ofs | (epl->len << ECS_pos_NBytes);
		epl->tr_ofs &= ~ECS_TogRESET;
		IntRestore(imf);
	}
}

#if WA_X_STALL_EP0_OUT
/*
 * permit hardware buffer to receive next OUT on EP0
 */
void uhw_arm_EP0_OUT_insteadof_STALL(void)
{
	struct _ep_list *epl;
	u_long imf;

	epl = &uhw_eplist[0];		/* fixed: EP0_OUT */

	imf = IntDisable();
	epl->len = epl->buff_size;
	//epl->tr_ofs &= ~ECS_TogRESET;		/* always unset for EP0_IN/EP0_OUT */
	*epl->list_at = ECS_ACTIVE | epl->tr_ofs | (epl->len << ECS_pos_NBytes);
	IntRestore(imf);
}
#endif /* WA_X_STALL_EP0_OUT */

/*
 * permit hardware to accept following token
 */
void uhw_ReadSETUP_finish(void)
{
	u_long imf;
	u_long dcs_base;

	if(LPC_USB->DEVCMDSTAT & DCS_SETUP) {
		imf = IntDisable();

		dcs_base = (LPC_USB->DEVCMDSTAT & ~DCS_0SAFEbits) | DCS_1SAFEbits;
		LPC_USB->DEVCMDSTAT = dcs_base | DCS_SETUP;

		IntRestore(imf);
	}
}

/*
 * convert 'bEndpointAddress' to LPC physical Endpoint number
 *   example: 0x00(EP0_OUT) ==> 0
 *            0x80(EP0_IN)  ==> 1
 *            0x01(EP1_OUT) ==> 2
 *            0x81(EP1_IN)  ==> 3
 */
static u_long get_EPnum(u_long EPaddr)
{
	return (EPaddr & 0x0F) * 2 + ((EPaddr >> 7) & 1);
}

/* =============================================================== */

/*
 * Handle interrupt from USB0
 */
void USB_IRQ_Handler(void)
{
	u_long intstat;
	u_long devcmdstat, dcs_base;
	int i;

SET_DEBUG_PORT;

#if myUDEBUG
//	PUTCHAR('I');
#endif /* myUDEBUG */

	intstat = LPC_USB->INTSTAT;
//RPRINTF("IS:%03x\r\n", (intstat >> 20) | (intstat & bits_EPall));
	intstat &= LPC_USB->INTEN;

	if(intstat & bitDEV_INT) {
		/*
		 * Device status interrupt
		 *   This bit is set by HW when one of the bits in the
		 *   Device Status Change register are set.
		 *   Software can clear this bit by writing a one to it.
		 */
		devcmdstat = LPC_USB->DEVCMDSTAT;
		dcs_base = (devcmdstat & ~DCS_0SAFEbits) | DCS_1SAFEbits;

		if(devcmdstat & DCS_DRES_C) {
			/*
			 * USB bus reset
			 *   On a bus reset the device will automatically go to the
			 *   default state (unconfigured and responding to address 0).
			 *        ... but ep-list memories are not modified,
			 *            so reset them by myself.
			 */
			struct _ep_list *epl;

			/*
			 * de-configure all endpoints by ep-list
			 */
			*uhw_eplist[0].list_at = ECS_enabled;	/* reset EP0_OUT */
			*uhw_eplist[1].list_at = ECS_enabled;	/* reset EP0_IN */
			i = 2;
			epl = &uhw_eplist[i];
			for(   ; i<USB_N_EP; i++) {		/* EP1_OUT, EP1_IN, ..., EP4_IN */
				*epl->list_at = ECS_DISABLED;
				epl++;
			}

			/*
			 * reset params on other layer
			 */
			GotSUD = FALSE;
			GotCOD = FALSE;
			Sleep = FALSE;
			ucore_ResetCore();
			if(USB_Jump_Table[USB_JT_RESET] != NULL) {
				USB_Jump_Table[USB_JT_RESET](0);
			}

			/*
			 * clear this bit only, keep others
			 */
			LPC_USB->DEVCMDSTAT = dcs_base | DCS_DRES_C;
		}

		if(devcmdstat & DCS_DSUS_C) {
			/*
			 * Suspend/Resume
			 */
			if(devcmdstat & DCS_DSUS) {
				/*
				 * Suspend
				 */
				GotSUD = FALSE;
				GotCOD = FALSE;
				Sleep = TRUE;
				ucore_SuspendCore();
				if(USB_Jump_Table[USB_JT_SUSPEND] != NULL) {
					USB_Jump_Table[USB_JT_SUSPEND](0);
//RPRINTF("DS:%08x\r\n", LPC_USB->DEVCMDSTAT);
				}
			} else {
				/*
				 * Resume
				 */
//				GotSUD = FALSE;
//				GotCOD = FALSE;
//				Sleep = FALSE;
				if(USB_Jump_Table[USB_JT_RESUME] != NULL) {
					USB_Jump_Table[USB_JT_RESUME](0);
				}
			}

			/*
			 * clear this bit only, keep others
			 */
			LPC_USB->DEVCMDSTAT = dcs_base | DCS_DSUS_C;
		}

		if(devcmdstat & DCS_DCON_C) {
			/*
			 * Connect change (Vbus voltage appeared/disappeared)
			 */
//			GotSUD = FALSE;
//			GotCOD = FALSE;
//			Sleep = FALSE;
			if(USB_Jump_Table[USB_JT_POWER] != NULL) {
				USB_Jump_Table[USB_JT_POWER](devcmdstat & DCS_VBUSDEBOUNCED);
			}

			/*
			 * clear this bit only, keep others
			 */
			LPC_USB->DEVCMDSTAT = dcs_base | DCS_DCON_C;
		}

		/*
		 * clear request
		 */
		LPC_USB->INTSTAT = bitDEV_INT;
	}

	if(intstat & bitFRM_INT) {
		/*
		 * Start of Frame Interrupt
		 */
		if(USB_Jump_Table[USB_JT_SOF] != NULL) {
			USB_Jump_Table[USB_JT_SOF]( LPC_USB->INFO & 0x7ff );
		}

		/*
		 * clear request
		 */
		LPC_USB->INTSTAT = bitFRM_INT;
	}

	if(intstat & bits_EPall) {
		/*
		 * Endpoint's Interrupt, Check All Endpoints
		 */
		for(i=0; i<USB_N_EP; i++) {
			if(intstat & (1<<i)) {
				/*
				 * clear request
				 */
				LPC_USB->INTSTAT = (1<<i);
	// ꂾONAȂ

				if(i == 0) {
					/* control out */
					if(LPC_USB->DEVCMDSTAT & DCS_SETUP) {
						/* this is Setup Packet, reset ep-list memories */
						*uhw_eplist[0].list_at = ECS_enabled;
						*uhw_eplist[1].list_at = ECS_enabled;
						GotSUD = TRUE;
						GotCOD = FALSE;
					} else {
						ucore_isr_EP0_out(0);
					}
				} else if(i == 1) {
					/* control in */
					ucore_isr_EP0_in(0);
				} else {
					/* generic Endpoints */
					if(USB_Jump_Table_EP[i] != NULL) {
						USB_Jump_Table_EP[i](0);
					}
				}
			}
		}
	}

CLR_DEBUG_PORT;
}

/* ######################################################## */
